from __future__ import annotations

import gc
from abc import ABC, abstractmethod
from contextlib import contextmanager
from pathlib import Path
from typing import TYPE_CHECKING, Any, Callable, Generator, Literal

import numpy as np
from funlib.geometry import Coordinate
from volara.utils import PydanticCoordinate, StrictBaseModel

if TYPE_CHECKING:
    import jax  # noqa: F401
    import torch


def _require_torch() -> None:
    try:
        import torch as _  # noqa: F811, F401
    except ImportError:
        raise ImportError(
            "PyTorch is required for this feature. "
            "Install it with: pip install volara-ml[torch]"
        ) from None


def _require_jax() -> None:
    try:
        import jax as _  # noqa: F811, F401
    except ImportError:
        raise ImportError(
            "JAX is required for this feature. "
            "Install it with: pip install volara-ml[jax]"
        ) from None


class Model(StrictBaseModel, ABC):
    """
    A base class for defining the common attributes and methods for all
    model types.
    """

    in_channels: int
    """
    The number of input channels to the model.
    """
    min_input_shape: PydanticCoordinate
    """
    The smallest possible input shape to the model. E.g. with UNets there
    is a minimum input size that satisfies the downsampling and all convolutional
    models have a minimum input size if using "valid" padding.
    """
    min_output_shape: PydanticCoordinate
    """
    The output shape to expect when providing the minimum input shape data.
    """
    min_step_shape: PydanticCoordinate
    """
    The minimum step size to increment the input shape by. This may be used
    to determine an optimal input/output shape for the model.
    """
    out_channels: list[int | None] | int | None
    """
    The number of channels in the output of the model. This can be an integer
    or list of integers if the model has multiple outputs. If `None` it is
    assumed that the number of output channels equals the number of input channels
    """
    out_range: tuple[float, float]
    """
    The range of the output data. This is used to convert between `np.uint8` and
    `np.float32` data types for efficient reading/writing of model outputs.
    """
    context_override: tuple[PydanticCoordinate, PydanticCoordinate] | None = None
    """
    An optional override for asymetrical context sizes. This lets you define
    a specific lower and upper context size for the model which must equal
    the expected context size of input_shape - output_shape
    """
    out_dtype: np.dtype = np.dtype(np.uint8)

    @property
    def context(self) -> Coordinate | tuple[Coordinate, Coordinate]:
        """
        The context required to make tile artifact free predictions
        with this model.
        """
        expected_context = self.min_input_shape - self.min_output_shape
        if self.context_override is not None:
            assert (
                self.context_override[0] + self.context_override[1]
            ) == expected_context, (
                f"Expected context override {self.context_override} to sum to {expected_context}, "
                f"but got {self.context_override[0] + self.context_override[1]}"
            )

        if self.context_override is not None:
            return self.context_override
        else:
            return expected_context // 2

    @abstractmethod
    def predict(self, device: Any) -> Generator[Callable[[np.ndarray], list[np.ndarray]], None, None]:
        """Context manager that yields a predict callable.

        Loads the model onto the given device on entry and frees GPU memory
        on exit (del model, clear caches, gc.collect).
        """
        pass

    @abstractmethod
    def select_device(self, worker_id: int) -> Any:
        """Return the device to use for the given worker."""
        pass

    @property
    @abstractmethod
    def eval_input_shape(self) -> Coordinate:
        """
        The input shape to use during prediction.
        """
        pass

    @property
    @abstractmethod
    def eval_output_shape(self) -> Coordinate:
        """
        The output shape to expect after providing data with shape
        `self.eval_input_shape()`
        """
        pass

    @property
    @abstractmethod
    def num_out_channels(self) -> list[int | None]:
        """
        The number of channels in each output prediction generated by this
        model. This is expected in a list to account for models that
        generate multiple output arrays.
        """
        pass

    def to_out_dtype(self, out_data: np.ndarray) -> np.ndarray:
        """
        A function defining the conversion function to go from model
        outputs to `np.uint8` for writing to zarr.
        """
        try:
            a_info = np.iinfo(self.out_dtype)
            a_min, a_max = a_info.min, a_info.max
        except ValueError:
            a_info = np.finfo(self.out_dtype)
            a_min, a_max = a_info.min, a_info.max
        return np.clip(
            ((out_data - self.out_range[0]) / (self.out_range[1] - self.out_range[0]))
            * a_max,
            a_min=a_min,
            a_max=a_max,
        ).astype(self.out_dtype)

    def from_uint8(self, data: np.ndarray) -> np.ndarray:
        """
        A function defining the conversion function to go from `np.uint8`
        back to `np.float32`
        """
        return (
            data.astype(np.float32) / 255 * (self.out_range[1] - self.out_range[0])
            + self.out_range[0]
        )


class TorchModel(Model):
    model_type: Literal["torch"] = "torch"

    def model_post_init(self, __context: object) -> None:
        _require_torch()

    save_path: Path
    """
    The path to a model saved using `torch.jit.save` or `torch.save`.
    """
    checkpoint_file: Path | None = None
    """
    An optional path to a checkpoint file containing the model weights.
    This can be provided either as a dictionary with key/value
    pair: "model_state_dict" and model.state_dict() or as a plain
    model.state_dict() dictionary."""
    pred_size_growth: PydanticCoordinate | None = None

    def model(self) -> "torch.nn.Module":
        import torch

        model = torch.load(self.save_path, map_location="cpu", weights_only=False)
        assert isinstance(model, torch.nn.Module), (
            f"Loading from {self.save_path} did not return a torch.nn.Module"
        )

        if self.checkpoint_file is not None:
            checkpoint = torch.load(
                self.checkpoint_file, map_location="cpu", weights_only=True
            )
            if "model_state_dict" in checkpoint:
                weights = checkpoint["model_state_dict"]
            else:
                weights = checkpoint
            model.load_state_dict(weights)

        return model

    @contextmanager
    def predict(self, device: Any) -> Generator[Callable[[np.ndarray], list[np.ndarray]], None, None]:
        import torch

        model = self.model()
        model.eval()
        model.to(device)

        def _predict(input_data: np.ndarray) -> list[np.ndarray]:
            with torch.no_grad():
                tensor = torch.as_tensor(input_data, device=device)
                out = model(tensor)  # noqa: F821
                if isinstance(out, tuple):
                    return [o.cpu().numpy() for o in out]
                return [out.cpu().numpy()]

        try:
            yield _predict
        finally:
            del model
            torch.cuda.empty_cache()
            gc.collect()

    def select_device(self, worker_id: int) -> Any:
        import torch

        n = torch.cuda.device_count()
        return "cpu" if n == 0 else f"cuda:{worker_id % n}"

    @property
    def eval_input_shape(self) -> Coordinate:
        input_shape = Coordinate(self.min_input_shape)

        if self.pred_size_growth is not None:
            assert np.sum(self.pred_size_growth % Coordinate(self.min_step_shape)) == 0
        if self.pred_size_growth is not None:
            input_shape = input_shape + self.pred_size_growth
        return input_shape

    @property
    def eval_output_shape(self) -> Coordinate:
        output_shape = Coordinate(self.min_output_shape)
        if self.pred_size_growth is not None:
            output_shape = output_shape + self.pred_size_growth
        return output_shape

    @property
    def num_out_channels(self) -> list[int | None]:
        num_channels = self.out_channels

        if isinstance(num_channels, int) or num_channels is None:
            return [num_channels]
        elif isinstance(num_channels, list):
            return num_channels


class JaxModel(Model):
    """A JAX/Flax model configuration.

    The model is loaded from a pickled Flax module (or apply function) and
    parameters are loaded from an orbax checkpoint directory or .msgpack file.
    All JAX imports are lazy so the module loads without JAX installed.
    """

    model_type: Literal["jax"] = "jax"
    model_path: Path
    """Path to a pickled Flax module (or apply function)."""
    params_path: Path | None = None
    """Path to an orbax checkpoint directory or .msgpack file with model parameters."""
    pred_size_growth: PydanticCoordinate | None = None

    def model_post_init(self, __context: object) -> None:
        _require_jax()

    def _load_model(self) -> Any:
        import pickle

        with open(self.model_path, "rb") as f:
            return pickle.load(f)

    def _load_params(self) -> Any:
        if self.params_path is None:
            return {}

        path = self.params_path
        if path.suffix == ".msgpack":
            from flax.serialization import from_bytes

            with open(path, "rb") as f:
                return from_bytes(None, f.read())
        else:
            import orbax.checkpoint as ocp

            checkpointer = ocp.StandardCheckpointer()
            return checkpointer.restore(path)

    @contextmanager
    def predict(self, device: Any) -> Generator[Callable[[np.ndarray], list[np.ndarray]], None, None]:
        import jax

        model = self._load_model()
        params = self._load_params()
        params = jax.device_put(params, device)

        @jax.jit
        def _apply(x: Any) -> Any:
            return model.apply(params, x)  # noqa: F821

        def _predict(input_data: np.ndarray) -> list[np.ndarray]:
            out = _apply(jax.numpy.asarray(input_data))
            if isinstance(out, tuple):
                return [np.asarray(o) for o in out]
            return [np.asarray(out)]

        try:
            yield _predict
        finally:
            del model, params
            jax.clear_caches()
            gc.collect()

    def select_device(self, worker_id: int) -> Any:
        import jax

        try:
            gpus = jax.devices("gpu")
        except RuntimeError:
            gpus = []
        if len(gpus) == 0:
            return jax.devices("cpu")[0]
        return gpus[worker_id % len(gpus)]

    @property
    def eval_input_shape(self) -> Coordinate:
        input_shape = Coordinate(self.min_input_shape)
        if self.pred_size_growth is not None:
            assert np.sum(self.pred_size_growth % Coordinate(self.min_step_shape)) == 0
            input_shape = input_shape + self.pred_size_growth
        return input_shape

    @property
    def eval_output_shape(self) -> Coordinate:
        output_shape = Coordinate(self.min_output_shape)
        if self.pred_size_growth is not None:
            output_shape = output_shape + self.pred_size_growth
        return output_shape

    @property
    def num_out_channels(self) -> list[int | None]:
        num_channels = self.out_channels
        if isinstance(num_channels, int) or num_channels is None:
            return [num_channels]
        elif isinstance(num_channels, list):
            return num_channels



import logging
from contextlib import contextmanager
from typing import Annotated, Callable, Literal

import daisy
import gunpowder as gp
import numpy as np
from funlib.geometry import Coordinate, Roi
from funlib.persistence import Array
from gunpowder import ArrayKey, Batch, BatchProvider
from gunpowder.nodes.generic_predict import GenericPredict
from pydantic import Field
from volara.blockwise import BlockwiseTask
from volara.datasets import LSD, Affs, Dataset, Raw

from ..models import JaxModel, Model, TorchModel

logger = logging.getLogger(__file__)


class ArraySource(BatchProvider):
    def __init__(self, key: ArrayKey, array: Array):
        self.key = key
        self.array = array

    def setup(self):
        spec = gp.ArraySpec(self.array.roi, self.array.voxel_size)
        self.provides(self.key, spec)

    def provide(self, request):
        outputs = Batch()
        data = self.array[request[self.key].roi]
        outputs[self.key] = gp.Array(
            data=data,
            spec=gp.ArraySpec(
                roi=request[self.key].roi,
                voxel_size=self.array.voxel_size,
            ),
        )
        return outputs


class ArrayWrite(gp.BatchFilter):
    def __init__(self, key: ArrayKey, array: Array, to_out_dtype: Callable):
        self.key = key
        self.array = array
        self.to_out_dtype = to_out_dtype

    def setup(self):
        self.updates(self.key, self.spec[self.key].copy())

    def process(self, batch, request):
        write_roi = request[self.key].roi.intersect(self.array.roi)
        data = batch[self.key].crop(write_roi).data
        self.array[write_roi] = self.to_out_dtype(data)


class CallablePredict(GenericPredict):
    """A framework-agnostic predict node that wraps a predict callable."""

    def __init__(
        self,
        predict_fn: Callable[[np.ndarray], list[np.ndarray]],
        input_key: ArrayKey,
        output_keys: dict[int, ArrayKey],
        spawn_subprocess: bool = False,
    ):
        self._predict_fn = predict_fn
        inputs = {0: input_key}
        super().__init__(inputs, output_keys, spawn_subprocess=spawn_subprocess)

    def predict(self, batch, request):
        input_data = batch[self.inputs[0]].data
        outputs = self._predict_fn(input_data)
        for idx, key in self.outputs.items():
            if key in request:
                spec = self.spec[key].copy()
                spec.roi = request[key].roi
                batch.arrays[key] = gp.Array(outputs[idx], spec)


OutDataType = Annotated[
    Raw | Affs | LSD,
    Field(discriminator="dataset_type"),
]


class Predict(BlockwiseTask):
    task_type: Literal["predict"] = "predict"
    checkpoint: Annotated[
        TorchModel | JaxModel,
        Field(discriminator="model_type"),
    ]
    in_data: Raw
    out_data: list[OutDataType | None]

    fit: Literal["overhang"] = "overhang"
    read_write_conflict: Literal[False] = False
    out_array_dtype: np.dtype = np.dtype(np.uint8)

    @property
    def checkpoint_config(self) -> Model:
        return self.checkpoint

    @property
    def write_roi(self) -> Roi:
        in_data_roi = self.in_data.array("r").roi
        if self.roi is not None:
            return in_data_roi.intersect(self.roi)
        else:
            return in_data_roi

    @property
    def voxel_size(self) -> Coordinate:
        return self.in_data.array("r").voxel_size

    @property
    def write_size(self) -> Coordinate:
        return self.checkpoint_config.eval_output_shape * self.voxel_size

    @property
    def context_size(self) -> Coordinate | tuple[Coordinate, Coordinate]:
        context = self.checkpoint_config.context
        if isinstance(context, Coordinate):
            return context * self.voxel_size
        elif isinstance(context[0], Coordinate) and isinstance(context[1], Coordinate):
            return (context[0] * self.voxel_size, context[1] * self.voxel_size)
        else:
            raise NotImplementedError(
                f"Unsupported context {context} type: {type(context)}. Expected Coordinate or tuple of Coordinates."
            )

    @property
    def task_name(self) -> str:
        return f"{'-'.join(dataset.name for dataset in self.output_datasets)}-{self.task_type}"

    @property
    def output_datasets(self) -> list[Dataset]:
        return [out_data for out_data in self.out_data if out_data is not None]

    def drop_artifacts(self):
        for out_data in self.out_data:
            if out_data is not None:
                out_data.drop()

    def init(self):
        self.init_out_array()

    def init_out_array(self):
        # TODO: hardcoding assuming channels first :/
        in_data = self.in_data.array("r")
        units = in_data.units
        axis_names = in_data.axis_names[-in_data.voxel_size.dims :]
        types = in_data.types[-in_data.voxel_size.dims :]
        for out_data, num_channels in zip(
            self.out_data, self.checkpoint_config.num_out_channels
        ):
            if num_channels is None:
                num_channels = self.in_data.array("r").shape[0]
            if out_data is not None:
                shape = (num_channels, *(self.write_roi.shape / self.voxel_size))
                chunk_shape = (num_channels, *self.write_size / self.voxel_size)
                out_data.prepare(
                    shape=shape,
                    chunk_shape=chunk_shape,
                    offset=self.write_roi.offset,
                    voxel_size=self.voxel_size,
                    units=units,
                    axis_names=[f"{out_data.name}^"] + axis_names,
                    types=[out_data.name] + types,
                    dtype=self.out_array_dtype,
                )

    @contextmanager
    def process_block_func(self):
        try:
            client = daisy.Client()
            device = self.checkpoint.select_device(client.worker_id)
        except KeyError:
            device = self.checkpoint.select_device(0)

        logging.info(f"using device {device}")

        input_key = gp.ArrayKey("INPUT_KEY")
        output_keys = [
            gp.ArrayKey(f"OUTPUT_KEY_{i}") for i in range(len(self.out_data))
        ]

        in_array = self.in_data.array("r")

        with self.checkpoint.predict(device) as predict_fn:
            pipeline = ArraySource(input_key, in_array)

            pipeline += gp.Pad(input_key, size=None)

            if in_array.channel_dims < 2:
                pipeline += gp.Stack(1)
            if in_array.channel_dims < 1:
                pipeline += gp.Stack(1)

            pipeline += CallablePredict(
                predict_fn=predict_fn,
                input_key=input_key,
                output_keys={
                    i: output_key
                    for i, output_key in enumerate(output_keys)
                    if self.out_data[i] is not None
                },
                spawn_subprocess=self.num_cache_workers is not None
                and self.num_cache_workers > 1,
            )

            pipeline += gp.Squeeze(
                [
                    output_key
                    for i, output_key in enumerate(output_keys)
                    if self.out_data[i] is not None
                ]
            )

            for output_key, out_data in zip(output_keys, self.out_data):
                if out_data is not None:
                    pipeline += ArrayWrite(
                        output_key, Dataset.array(out_data, "a"), self.checkpoint.to_out_dtype
                    )

            print("Starting prediction...")

            with gp.build(pipeline):

                def process_block(block):
                    request = gp.BatchRequest()
                    request[input_key] = gp.ArraySpec(roi=block.read_roi)
                    for i, output_key in enumerate(output_keys):
                        if self.out_data[i] is not None:
                            request[output_key] = gp.ArraySpec(roi=block.write_roi)
                    pipeline.request_batch(request)

                yield process_block



import cloudpickle
import jax
import jax.numpy as jnp
import numpy as np
import orbax.checkpoint as ocp
import pytest
import torch
from flax import linen as nn
from funlib.geometry import Coordinate

from volara_ml.models import JaxModel, TorchModel


@pytest.mark.parametrize("save_type", ["jit", "pickle"])
@pytest.mark.parametrize("out_range", [(0, 1), (-1, 1)])
@pytest.mark.parametrize("checkpoint", [True, False])
def test_torch_models(save_type, out_range, checkpoint, tmp_path):
    test_data = torch.randn(1, 1, 100, 100)
    model = torch.nn.Sequential(
        torch.nn.Conv2d(1, 1, 3, padding="same"), torch.nn.Sigmoid()
    )
    test_output = model(test_data)
    weights = model.state_dict()
    checkpoint_file = tmp_path / "checkpoint.pth"
    torch.save({"model_state_dict": weights}, checkpoint_file)
    model = torch.nn.Sequential(
        torch.nn.Conv2d(1, 1, 3, padding="same"), torch.nn.Sigmoid()
    )
    test_output2 = model(test_data)
    assert not torch.allclose(test_output, test_output2)
    if save_type == "jit":
        save_path = tmp_path / "model.jit"
        trace = torch.jit.trace(model, test_data)
        torch.jit.save(trace, save_path)
    elif save_type == "pickle":
        save_path = tmp_path / "model.pth"
        torch.save(model, save_path)

    model_config = TorchModel(
        in_channels=1,
        min_input_shape=Coordinate(1, 1),
        min_output_shape=Coordinate(1, 1),
        min_step_shape=Coordinate(1, 1),
        out_channels=1,
        out_range=out_range,
        save_path=save_path,
        checkpoint_file=checkpoint_file if checkpoint else None,
        pred_size_growth=Coordinate(99, 99),
    )

    # basic model interface values
    assert model_config.context == (0, 0)
    assert model_config.eval_input_shape == (100, 100)
    assert model_config.eval_output_shape == (100, 100)
    assert model_config.num_out_channels == [1]

    # model data conversions
    uint8_data = np.arange(0, 256, dtype=np.uint8).reshape(16, 16)
    float_data = model_config.from_uint8(uint8_data)
    uint8_data_back = model_config.to_out_dtype(float_data)
    assert np.isclose(float_data.max(), model_config.out_range[1])
    assert np.isclose(float_data.min(), model_config.out_range[0])
    assert uint8_data_back.max() == 255, (
        uint8_data.min(),
        uint8_data.max(),
        float_data.min(),
        float_data.max(),
        uint8_data_back.min(),
        uint8_data_back.max(),
    )
    assert uint8_data_back.min() == 0

    # test model runs as expected
    if checkpoint:
        assert torch.allclose(model_config.model()(test_data), test_output)
    else:
        assert torch.allclose(model_config.model()(test_data), test_output2)


class Identity(nn.Module):
    @nn.compact
    def __call__(self, x):
        x = jnp.transpose(x, (0, 2, 3, 1))
        x = nn.Conv(
            features=1,
            kernel_size=(1, 1),
            use_bias=False,
            kernel_init=nn.initializers.ones,
        )(x)
        x = jnp.transpose(x, (0, 3, 1, 2))
        return x


@pytest.mark.parametrize("save_type", ["orbax", "msgpack"])
@pytest.mark.parametrize("out_range", [(0, 1), (-1, 1)])
def test_jax_models(save_type, out_range, tmp_path):
    model = Identity()
    rng = jax.random.PRNGKey(0)
    test_data = jax.random.normal(rng, (1, 1, 100, 100))
    params = model.init(rng, test_data)
    test_output = model.apply(params, test_data)

    # Save model via cloudpickle
    model_path = tmp_path / "model.pkl"
    with open(model_path, "wb") as f:
        cloudpickle.dump(model, f)

    # Save params
    if save_type == "orbax":
        params_path = tmp_path / "ckpt"
        checkpointer = ocp.StandardCheckpointer()
        checkpointer.save(str(params_path), params)
        checkpointer.close()
    elif save_type == "msgpack":
        from flax.serialization import to_bytes

        params_path = tmp_path / "params.msgpack"
        with open(params_path, "wb") as f:
            f.write(to_bytes(params))

    model_config = JaxModel(
        in_channels=1,
        min_input_shape=Coordinate(1, 1),
        min_output_shape=Coordinate(1, 1),
        min_step_shape=Coordinate(1, 1),
        out_channels=1,
        out_range=out_range,
        model_path=model_path,
        params_path=params_path,
        pred_size_growth=Coordinate(99, 99),
    )

    # basic model interface values
    assert model_config.context == (0, 0)
    assert model_config.eval_input_shape == (100, 100)
    assert model_config.eval_output_shape == (100, 100)
    assert model_config.num_out_channels == [1]

    # model data conversions
    uint8_data = np.arange(0, 256, dtype=np.uint8).reshape(16, 16)
    float_data = model_config.from_uint8(uint8_data)
    uint8_data_back = model_config.to_out_dtype(float_data)
    assert np.isclose(float_data.max(), model_config.out_range[1])
    assert np.isclose(float_data.min(), model_config.out_range[0])
    assert uint8_data_back.max() == 255, (
        uint8_data.min(),
        uint8_data.max(),
        float_data.min(),
        float_data.max(),
        uint8_data_back.min(),
        uint8_data_back.max(),
    )
    assert uint8_data_back.min() == 0

    # test model runs via predict context manager
    device = model_config.select_device(0)
    with model_config.predict(device) as predict_fn:
        result = predict_fn(np.array(test_data))
        assert len(result) == 1
        np.testing.assert_allclose(result[0], np.array(test_output), atol=1e-5)
